import configparser

from os.path import expanduser
from typing import Tuple


class MissingCredentials(Exception):
    pass


class Credentials:
    """Used to fetch and store credentials. One of 3 conditions must be met to successfully create credentials.

    1. credentials_path is provided
    2. access_key_id, secret_access_key and api_key is provided
    3. credentials is located in default path ~/.lucidtech/credentials.cfg

    :param credentials_path: Path to credentials file
    :type credentials_path: str
    :param access_key_id: Access Key Id
    :type access_key_id: str
    :param secret_access_key: Secret Access Key
    :type secret_access_key: str
    :param api_key: API key
    :type api_key: str

    """
    def __init__(self, credentials_path=None, access_key_id=None, secret_access_key=None, api_key=None):
        if any([not access_key_id, not secret_access_key, not api_key]):
            if not credentials_path:
                credentials_path = expanduser('~/.lucidtech/credentials.cfg')

            access_key_id, secret_access_key, api_key = self._read_credentials(credentials_path)

        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key
        self.api_key = api_key

        if any([not access_key_id, not secret_access_key, not api_key]):
            raise MissingCredentials

    @staticmethod
    def _read_credentials(credentials_path: str) -> Tuple[str, str, str]:
        config = configparser.ConfigParser()
        config.read(credentials_path)
        section = 'default'

        access_key_id = config.get(section, 'access_key_id')
        secret_access_key = config.get(section, 'secret_access_key')
        api_key = config.get(section, 'api_key')

        return access_key_id, secret_access_key, api_key



import requests
import json
import pathlib
import logging

from backoff import on_exception, expo
from requests.exceptions import RequestException
from json.decoder import JSONDecodeError
from urllib.parse import urlparse
from typing import List, Dict

from .authorization import Authorization
from .credentials import Credentials


logger = logging.getLogger('las')


def _fatal_code(e):
    return 400 <= e.response.status_code < 500


def _json_decode(response):
    try:
        response.raise_for_status()
        return response.json()
    except JSONDecodeError as e:
        logger.error('Error in response. Returned {}'.format(response.text))
        raise e
    except Exception as e:
        logger.error('Error in response. Returned {}'.format(response.text))

        if response.status_code == 403 and 'Forbidden' in response.json().values():
            raise InvalidCredentialsException('Credentials provided is not valid')

        raise e


class ClientException(Exception):
    """A ClientException is raised if the client refuses to
    send request due to incorrect usage or bad request data."""
    pass


class InvalidCredentialsException(ClientException):
    """An InvalidCredentialsException is raised if api key, access key id or secret access key is invalid."""
    pass


class Client:
    """A low level client to invoke api methods from Lucidtech AI Services.

    :param endpoint: Domain endpoint of the api, e.g. https://<prefix>.api.lucidtech.ai/<version>
    :type endpoint: str
    :param credentials: Credentials to use, instance of :py:class:`~las.Credentials`
    :type credentials: Credentials

    """
    def __init__(self, endpoint: str, credentials=None):
        self.endpoint = endpoint
        credentials = credentials or Credentials()
        self.authorization = Authorization(credentials)

    @on_exception(expo, RequestException, max_tries=3, giveup=_fatal_code)
    def post_documents(self, content_type: str, consent_id: str) -> dict:
        """Creates a document handle, calls the POST /documents endpoint.

        >>> from las import Client
        >>> client = Client(endpoint='<api endpoint>')
        >>> client.post_documents('image/jpeg', consent_id='foobar')

        :param content_type: A mime type for the document handle
        :type content_type: str
        :param consent_id: An identifier to mark the owner of the document handle
        :type consent_id: str
        :return: Document handle id and pre-signed upload url
        :rtype: dict
        :raises InvalidCredentialsException: If the credentials are invalid
        :raises requests.exception.RequestException: If error was raised by requests
        """

        body = json.dumps({'contentType': content_type, 'consentId': consent_id}).encode()
        uri, headers = self._create_signing_headers('POST', '/documents', body)

        post_documents_response = requests.post(
            url=uri.geturl(),
            headers=headers,
            data=body
        )

        response = _json_decode(post_documents_response)
        return response

    @staticmethod
    @on_exception(expo, RequestException, max_tries=3, giveup=_fatal_code)
    def put_document(document_path: str, content_type: str, presigned_url: str) -> str:
        """Convenience method for putting a document to presigned url.

        >>> from las import Client
        >>> client = Client(endpoint='<api endpoint>')
        >>> client.put_document(document_path='document.jpeg', content_type='image/jpeg', presigned_url='<presigned url>')

        :param document_path: Path to document to upload
        :type document_path: str
        :param content_type: Mime type of document to upload. Same as provided to :py:func:`~las.Client.post_documents`
        :type content_type: str
        :param presigned_url: Presigned upload url from :py:func:`~las.Client.post_documents`
        :type presigned_url: str
        :return: Response from put operation
        :rtype: str
        :raises requests.exception.RequestException: If error was raised by requests
        """

        body = pathlib.Path(document_path).read_bytes()
        headers = {'Content-Type': content_type}
        put_document_response = requests.put(presigned_url, data=body, headers=headers)
        put_document_response.raise_for_status()
        return put_document_response.content.decode()

    @on_exception(expo, RequestException, max_tries=3, giveup=_fatal_code)
    def post_predictions(self, document_id: str, model_name: str) -> dict:
        """Run inference and create a prediction, calls the POST /predictions endpoint.

        >>> from las import Client
        >>> client = Client(endpoint='<api endpoint>')
        >>> client.post_predictions(document_id='<document id>', model_name='invoice')

        :param document_id: The document id to run inference and create a prediction on
        :type document_id: str
        :param model_name: The name of the model to use for inference
        :type model_name: str
        :return: Prediction on document
        :rtype: dict
        :raises InvalidCredentialsException: If the credentials are invalid
        :raises requests.exception.RequestException: If error was raised by requests
        """

        body = json.dumps({'documentId': document_id, 'modelName': model_name}).encode()
        uri, headers = self._create_signing_headers('POST', '/predictions', body)

        post_predictions_response = requests.post(
            url=uri.geturl(),
            headers=headers,
            data=body
        )

        response = _json_decode(post_predictions_response)
        return response

    @on_exception(expo, RequestException, max_tries=3, giveup=_fatal_code)
    def post_document_id(self, document_id: str, feedback: List[Dict[str, str]]) -> dict:
        """Post feedback to the REST API, calls the POST /documents/{documentId} endpoint.
        Posting feedback means posting the ground truth data for the particular document.
        This enables the API to learn from past mistakes.

        >>> from las import Client
        >>> client = Client(endpoint='<api endpoint>')
        >>> feedback = [{'label': 'total_amount', 'value': '156.00'}, {'label': 'invoice_date', 'value': '2018-10-23'}]
        >>> client.post_document_id(document_id='<document id>', feedback=feedback)

        :param document_id: The document id to run inference and create a prediction on
        :type document_id: str
        :param feedback: A list of feedback items
        :type feedback: List[Dict[str, str]]
        :return: Feedback response from REST API
        :rtype: dict
        :raises InvalidCredentialsException: If the credentials are invalid
        :raises requests.exception.RequestException: If error was raised by requests
        """

        body = json.dumps({'feedback': feedback}).encode()
        uri, headers = self._create_signing_headers('POST', f'/documents/{document_id}', body)

        post_document_id_response = requests.post(
            url=uri.geturl(),
            headers=headers,
            data=body
        )

        response = _json_decode(post_document_id_response)
        return response

    @on_exception(expo, RequestException, max_tries=3, giveup=_fatal_code)
    def delete_consent_id(self, consent_id: str) -> dict:
        """Delete documents with this consent_id, calls the DELETE /consent/{consentId} endpoint.

        >>> from las import Client
        >>> client = Client(endpoint='<api endpoint>')
        >>> client.delete_consent_id('<consent id>')

        :param consent_id: Delete documents with this consent_id
        :type consent_id: str
        :return: Delete consent id response from REST API
        :rtype: dict
        :raises InvalidCredentialsException: If the credentials are invalid
        :raises requests.exception.RequestException: If error was raised by requests
        """

        body = json.dumps({}).encode()
        uri, headers = self._create_signing_headers('DELETE', f'/consents/{consent_id}', body)

        delete_consent_id_consent = requests.delete(
            url=uri.geturl(),
            headers=headers,
            data=body
        )

        response = _json_decode(delete_consent_id_consent)
        return response

    def _create_signing_headers(self, method: str, path: str, body: bytes):
        uri = urlparse(f'{self.endpoint}{path}')

        auth_headers = self.authorization.sign_headers(
            uri=uri,
            method=method,
            body=body
        )

        headers = {**auth_headers, 'Content-Type': 'application/json'}
        return uri, headers



class Field(dict):
    def __init__(self, label, value, confidence=None):
        field = dict(label=label, value=value)
        field = dict(**field, confidence=confidence) if confidence else field
        super().__init__(**field)

    @property
    def label(self):
        return self['label']

    @property
    def value(self):
        return self['value']

    @property
    def confidence(self):
        return self.get('confidence')


class Prediction(dict):
    def __init__(self, document_id: str, consent_id: str, model_name: str, prediction_response: dict):
        prediction = dict(
            document_id=document_id,
            consent_id=consent_id,
            model_name=model_name,
            fields=prediction_response['predictions']
        )

        super().__init__(**prediction)

    @property
    def document_id(self):
        return self['document_id']

    @property
    def consent_id(self):
        return self['consent_id']

    @property
    def model_name(self):
        return self['model_name']

    @property
    def fields(self):
        return [Field(**field) for field in self['fields']]



import pathlib
import imghdr
import logging

from uuid import uuid4
from typing import List

from ._extrahdr import extra_what
from .client import Client
from .prediction import Prediction, Field


logger = logging.getLogger('las')


class FileFormatException(Client):
    """A FileFormatException is raised if the file format is not supported by the api."""
    pass


class ApiClient(Client):
    """A high level client to invoke api methods from Lucidtech AI Services.

    :param endpoint: Domain endpoint of the api, e.g. https://<prefix>.api.lucidtech.ai/<version>
    :type endpoint: str
    :param credentials: Credentials to use, instance of :py:class:`~las.Credentials`
    :type credentials: Credentials

    """
    def predict(self, document_path: str, model_name: str, consent_id: str = None) -> Prediction:
        """Run inference and create prediction on document.
        This method takes care of creating and uploading a document specified by document_path.
        as well as running inference using model specified by model_name to create prediction on the document.

        >>> from las import ApiClient
        >>> api_client = ApiClient(endpoint='<api endpoint>')
        >>> api_client.predict(document_path='document.jpeg', model_name='invoice')

        :param document_path: Path to document to run inference on
        :type document_path: str
        :param model_name: The name of the model to use for inference
        :type model_name: str
        :param consent_id: An identifier to mark the owner of the document handle
        :type consent_id: str
        :return: Prediction on document
        :rtype: Prediction
        :raises InvalidCredentialsException: If the credentials are invalid
        :raises requests.exception.RequestException: If error was raised by requests
        """

        content_type = self._get_content_type(document_path)
        consent_id = consent_id or str(uuid4())
        document_id = self._upload_document(document_path, content_type, consent_id)
        prediction_response = self.post_predictions(document_id, model_name)
        return Prediction(document_id, consent_id, model_name, prediction_response)

    def send_feedback(self, document_id: str, feedback: List[Field]) -> dict:
        """Send feedback to the model.
        This method takes care of sending feedback related to document specified by document_id.
        Feedback consists of ground truth values for the document specified as a list of Field instances.

        >>> from las import ApiClient
        >>> api_client = ApiClient(endpoint='<api endpoint>')
        >>> feedback = [Field(label='total_amount', value='120.00'), Field(label='purchase_date', value='2019-03-10')]
        >>> api_client.send_feedback('<document id>', feedback)

        :param document_id: The document id of the document that will receive the feedback
        :type document_id: str
        :param feedback: A list of :py:class:`~las.Field` representing the ground truth values for the document
        :type feedback: List[Field]
        :return: Feedback response
        :rtype: dict
        :raises InvalidCredentialsException: If the credentials are invalid
        :raises requests.exception.RequestException: If error was raised by requests
        """

        return self.post_document_id(document_id, feedback)

    def revoke_consent(self, consent_id: str) -> dict:
        """Revoke consent and deleting all documents associated with consent_id.
        Consent id is a parameter that is provided by the user upon making a prediction on a document.
        See :py:func: `~las.ApiClient.predict`.

        >>> from las import ApiClient
        >>> api_client = ApiClient(endpoint='<api endpoint>')
        >>> api_client.revoke_consent('<consent id>')

        :param consent_id: Delete documents associated with this consent_id
        :type consent_id: str
        :return: Revoke consent response
        :rtype: dict
        :raises InvalidCredentialsException: If the credentials are invalid
        :raises requests.exception.RequestException: If error was raised by requests
        """
        return self.delete_consent_id(consent_id)

    def _upload_document(self, document_path: str, content_type: str, consent_id: str) -> str:
        post_documents_response = self.post_documents(content_type, consent_id)
        document_id = post_documents_response['documentId']
        presigned_url = post_documents_response['uploadUrl']
        self.put_document(document_path, content_type, presigned_url)
        return document_id

    @staticmethod
    def _get_content_type(document_path: str) -> str:
        supported_formats = {
            'jpeg': 'image/jpeg',
            'pdf': 'application/pdf'
        }

        fp = pathlib.Path(document_path)
        fmt = imghdr.what(fp) or extra_what(fp)

        if fmt in supported_formats:
            return supported_formats[fmt]
        else:
            raise FileFormatException



import logging

from .api_client import ApiClient
from .client import Client
from .credentials import Credentials
from .prediction import Prediction, Field


logger = logging.getLogger('las')
logger.addHandler(logging.NullHandler())
logger.setLevel(logging.INFO)



import datetime
import hashlib
import hmac

from urllib.parse import ParseResult
from typing import Tuple, Iterable

from .credentials import Credentials


class Authorization:
    ALGORITHM = 'AWS4-HMAC-SHA256'

    def __init__(self, credentials: Credentials):
        self.region = 'eu-west-1'
        self.service = 'execute-api'
        self.credentials = credentials

    def sign_headers(self, uri: ParseResult, method: str, body: bytes = None) -> dict:
        if not body:
            body = ''.encode('utf-8')

        amz_date, datestamp = self._timestamp()

        canonical_request, headers = self._get_canonical_request(
            uri=uri,
            method=method,
            body=body,
            amz_date=amz_date)

        req_digest = hashlib.sha256(canonical_request).hexdigest()
        cred_scope = self._get_credential_scope(datestamp)
        string_to_sign = self._get_sign_string(amz_date, cred_scope, req_digest)
        signing_key = self._get_signature_key(datestamp)

        signature = hmac.new(signing_key, string_to_sign, hashlib.sha256).hexdigest()

        auth_header = self._build_auth_header(
            amz_date=amz_date,
            access_key=self.credentials.access_key_id,
            api_key=self.credentials.api_key,
            signature=signature,
            credential_scope=cred_scope,
            signed_headers=headers
        )

        return auth_header

    def _get_sign_string(self, amz_date: str, cred_scope: str, req_digest: str) -> bytes:
        sign_parts = [self.ALGORITHM, amz_date, cred_scope, req_digest]
        return '\n'.join(sign_parts).encode('utf-8')

    def _get_credential_scope(self, datestamp: str) -> str:
        scope_parts = [datestamp, self.region, self.service, 'aws4_request']
        return '/'.join(scope_parts)

    def _get_canonical_request(self, uri: ParseResult, method: str, body: bytes, amz_date: str) \
            -> Tuple[bytes, Iterable[str]]:
        canonical_querystring = self._get_canonical_querystring(uri.query)

        header_keys = ('host', 'x-amz-date', 'x-api-key')
        header_values = (uri.netloc, amz_date, self.credentials.api_key)
        headers = dict(zip(header_keys, header_values))

        header_parts = (f'{k}:{v}\n' for k, v in headers.items())
        payload_hash = hashlib.sha256(body).hexdigest()

        request_components = [
            method, uri.path, canonical_querystring,
            ''.join(header_parts), ';'.join(headers),
            payload_hash]

        return '\n'.join(request_components).encode('utf-8'), headers.keys()

    # TODO: Implement AWS sig v4 canonical querystring
    def _get_canonical_querystring(self, querystring: str) -> str:
        if querystring:
            raise NotImplementedError('Creating canonical querystring is not implemented')
        return ''

    def _get_signature_key(self, datestamp: str) -> str:
        def sign(key, msg):
            return hmac.new(key, msg.encode('utf-8'), hashlib.sha256).digest()

        signature = ('AWS4' + self.credentials.secret_access_key).encode('utf-8')
        for part in (datestamp, self.region, self.service, 'aws4_request'):
            signature = sign(signature, part)

        return signature

    @classmethod
    def _build_auth_header(cls, amz_date: str, access_key: str, api_key: str, signature: str,
                           credential_scope: str, signed_headers: Iterable[str]) -> dict:
        auth = {
            'Credential': f'{access_key}/{credential_scope}',
            'SignedHeaders': ';'.join(signed_headers),
            'Signature': signature
        }

        auth_parts = (f'{k}={auth[k]}' for k in auth)
        auth_string = ', '.join(auth_parts)

        headers = {
            'x-amz-date': amz_date,
            'x-api-key': api_key,
            'Authorization': f'{cls.ALGORITHM} {auth_string}'
        }

        headers = {k: v for k, v in headers.items() if v}
        return headers

    @staticmethod
    def _timestamp() -> Tuple[str, str]:
        now = datetime.datetime.utcnow()
        amz_date = now.strftime('%Y%m%dT%H%M%SZ')
        datestamp = now.strftime('%Y%m%d')
        return amz_date, datestamp
